;;; acme-mode.el --- mode for editing assembler code for ACME assembler
;;; based on emacs asm-mode.el by Eric S. Raymond
;;; written 2020 by Ralf Hoffmann <ralf@boomerangsworld.de>
;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Version 2020-03-16

(defgroup acme nil
  "Mode for editing ACME assembler code."
  :link '(custom-group-link :tag "Font Lock Faces group" font-lock-faces)
  :group 'languages)

(defcustom acme-comment-char ?\;
  "The comment-start character assumed by Acme mode."
  :type 'character
  :group 'acme)

(defvar acme-mode-syntax-table
  (let ((st (make-syntax-table)))
    (modify-syntax-entry ?\n "> b" st)
    st)
  "Syntax table used while in Acme mode.")

(defvar acme-mode-abbrev-table nil
  "Abbrev table used while in Acme mode.")
(define-abbrev-table 'acme-mode-abbrev-table ())

(defvar acme-mode-map
  (let ((map (make-sparse-keymap)))
    ;; Note that the comment character isn't set up until acme-mode is called.
    (define-key map ":"		'acme-colon)
    (define-key map "+"		(lambda () (interactive) (acme-anon-label "+")))
    (define-key map "-"		(lambda () (interactive) (acme-anon-label "-")))
    ;(define-key map "\t"		'acme-indent-and-tab)
    (define-key map (kbd "<backtab>")		'acme-indent-line)
    (define-key map "\C-c;"	'comment-region)
    (define-key map "\C-j"	'newline-and-indent)
    (define-key map "\C-m"	'newline-and-indent)
    (define-key map [menu-bar acme-mode] (cons "Acme" (make-sparse-keymap)))
    (define-key map [menu-bar acme-mode comment-region]
      '(menu-item "Comment Region" comment-region
		  :help "Comment or uncomment each line in the region"))
    (define-key map [menu-bar acme-mode newline-and-indent]
      '(menu-item "Insert Newline and Indent" newline-and-indent
		  :help "Insert a newline, then indent according to major mode"))
    (define-key map [menu-bar acme-mode acme-colon]
      '(menu-item "Insert Colon" acme-colon
		  :help "Insert a colon; if it follows a label, delete the label's indentation"))
    map)
  "Keymap for Acme mode.")

(defconst acme-font-lock-keywords
  (append
   '(("^\\(\\(\\sw\\|\\s_\\)+\\)\\>:?[ \t]*\\(\\sw+\\(\\.\\sw+\\)*\\)?"
      (1 font-lock-function-name-face) (3 font-lock-keyword-face nil t))
     ;; label started from ".".
     ("^\\([\\@\\.]\\(\\sw\\|\\s_\\)+\\)\\>:?[ \t]*\\(\\sw+\\(\\.\\sw+\\)*\\)?"
      (1 font-lock-function-name-face) (3 font-lock-keyword-face nil t))
     ;; anonymous label
     ("^\\([-+]+\\)\\([ \t]+\\(\\sw+\\(\\.\\sw+\\)*\\)\\)?"
      (1 font-lock-function-name-face) (2 font-lock-keyword-face nil t))
     ("^\\((\\sw+)\\)?\\s +\\(\\(\\.?\\sw\\|\\s_\\)+\\(\\.\\sw+\\)*\\)"
      2 font-lock-keyword-face)
     ;; directive started from ".".
     ("^\\(\\.\\(\\sw\\|\\s_\\)+\\)\\>[^:]?"
      1 font-lock-keyword-face)
     ;; %register
     ("%\\sw+" . font-lock-variable-name-face)
     ("[#!]\\(\\sw\\|\\s_\\)+" . font-lock-variable-name-face)
    ;("$[A-Fa-f0-9]+" . font-lock-function-name-face)
     ("\\b\\([$0-9][0-9a-fA-F]*\\)" . font-lock-function-name-face)
     (",[xy]" . font-lock-keyword-face))
   cpp-font-lock-keywords)
  "Additional expressions to highlight in Assembler mode.")

(defun acme-in-comment-paragraph (body)
  ;; We might move the point to fill the next comment, but we don't want it
  ;; seeming to jump around on the user
  (save-excursion
    ;; If we're outside of a comment, with only whitespace and then a comment
    ;; in front, jump to the comment and prepare to fill it.
    (when (not (nth 4 (syntax-ppss)))
      (beginning-of-line)
      (when (looking-at (concat "[[:space:]\n]*" comment-start-skip))
        (goto-char (match-end 0))))

    ;; We need this when we're moving the point around and then checking syntax
    ;; while doing paragraph fills, because the cache it uses isn't always
    ;; invalidated during this.
    (syntax-ppss-flush-cache 1)
    ;; If we're at the beginning of a comment paragraph with nothing but
    ;; whitespace til the next line, jump to the next line so that we use the
    ;; existing prefix to figure out what the new prefix should be, rather than
    ;; inferring it from the comment start.
    (let ((next-bol (line-beginning-position 2)))
      (while (save-excursion
               (end-of-line)
               (syntax-ppss-flush-cache 1)
               (and (nth 4 (syntax-ppss))
                    (save-excursion
                      (beginning-of-line)
                      (looking-at paragraph-start))
                    (looking-at "[[:space:]]*$")
                    (nth 4 (syntax-ppss next-bol))))
        (goto-char next-bol)))

    (syntax-ppss-flush-cache 1)
    ;; If we're on the last line of a multiline-style comment that started
    ;; above, back up one line so we don't mistake the * of the */ that ends
    ;; the comment for a prefix.
    (when (save-excursion
            (and (nth 4 (syntax-ppss (line-beginning-position 1)))
                 (looking-at "[[:space:]]*\\;")))
      (goto-char (line-end-position 0)))
    (funcall body)))

(defun clean-line (line)
  ;; replace every none tab character by a space to prepare for
  ;; indentation
  (if (= 0 (length line))
      ""
    (let ((first (substring line 0 1))
          (rest (substring line 1)))
      (if (string= first "\t")
          (concat first (clean-line rest))
        (concat " " (clean-line rest))))))

(defun acme-with-comment-fill-prefix (body)
  (let*
      ((line-string (buffer-substring-no-properties
                     (line-beginning-position) (line-end-position)))
       (line-comment-start
        (when (nth 4 (syntax-ppss))
          (cond
           ;; If we're inside the comment and see a ; prefix, use it
           ((string-match "^\\([[:space:]]*\\;+[[:space:]]*\\)"
                          line-string)
            (match-string 1 line-string))
           ;; If we're at the start of a comment, figure out what prefix
           ;; to use for the subsequent lines after it
           ((string-match (concat "^\\(.*[[:space:]]*\\)" comment-start-skip) line-string)
            (concat (clean-line (match-string 1 line-string)) comment-start)))))
       (fill-prefix
        (or line-comment-start
            fill-prefix)))
    (funcall body)))

(defun acme-fill-paragraph (&rest args)
  "Special wrapping for `fill-paragraph' to handle multi-line comments with a * prefix on each line."
  (acme-in-comment-paragraph
   (lambda ()
     (acme-with-comment-fill-prefix
      (lambda ()
        (let
            ((fill-paragraph-function
              (if (not (eq fill-paragraph-function 'acme-fill-paragraph))
                  fill-paragraph-function))
             (fill-paragraph-handle-comment t))
          (apply 'fill-paragraph args)
          t))))))

(defun acme-fill-forward-paragraph (arg)
  ;; This is to work around some funny behavior when a paragraph separator is
  ;; at the very top of the file and there is a fill prefix.
  (let ((fill-prefix nil)) (forward-paragraph arg)))

(defun acme-comment-indent-new-line (&optional arg)
  (acme-with-comment-fill-prefix
   (lambda () (comment-indent-new-line arg))))

;;;###autoload
(add-to-list 'auto-mode-alist '("\\.acme\\'" . acme-mode))

(define-derived-mode acme-mode prog-mode "ACME"
  "Major mode for editing ACME assembler code.
Features a private abbrev table and the following bindings:

\\[acme-colon]\toutdent a preceding label, tab to next tab stop.
\\[tab-to-tab-stop]\ttab to next tab stop.
\\[acme-newline]\tnewline, then tab to next tab stop.
\\[acme-comment]\tsmart placement of assembler comments.

The character used for making comments is set by the variable
`acme-comment-char' (which defaults to `?\\;').

Alternatively, you may set this variable in `acme-mode-set-comment-hook',
which is called near the beginning of mode initialization.

Turning on Acme mode runs the hook `acme-mode-hook' at the end of initialization.

Special commands:
\\{acme-mode-map}"
  (setq local-abbrev-table acme-mode-abbrev-table)
  (set (make-local-variable 'font-lock-defaults) '(acme-font-lock-keywords))
  (set (make-local-variable 'indent-line-function) 'acme-indent-line)
  ;; Stay closer to the old TAB behavior (was tab-to-tab-stop).
  (set (make-local-variable 'tab-always-indent) nil)
  (setq-local tab-width 8)
  (setq-local comment-multi-line t)
  (setq-local comment-start (concat (char-to-string acme-comment-char) " "))
  (setq-local comment-end   "")
  (setq-local indent-tabs-mode t)
  (setq-local fill-paragraph-function 'acme-fill-paragraph)
  (setq-local fill-forward-paragraph-function 'acme-fill-forward-paragraph)
  (setq-local comment-line-break-function 'acme-comment-indent-new-line)
  (setq-local comment-column 48)

  (setq-local
   electric-indent-chars (cons ?} (and (boundp 'electric-indent-chars)
                                       electric-indent-chars)))

  (run-hooks 'acme-mode-set-comment-hook)
  ;; Make our own local child of acme-mode-map
  ;; so we can define our own comment character.
  (use-local-map (nconc (make-sparse-keymap) acme-mode-map))
  (local-set-key (vector acme-comment-char) 'acme-comment)
  (set-syntax-table (make-syntax-table acme-mode-syntax-table))
  (modify-syntax-entry	acme-comment-char "< b")

  (setq cs (regexp-quote (char-to-string acme-comment-char)))
  (set (make-local-variable 'comment-add) 1)
  (set (make-local-variable 'comment-start-skip)
       (concat cs "+[ \t]*"))
  (set (make-local-variable 'comment-end-skip) "[ \t]*\\(\\s>\\|\\*+/\\)")
  (setq fill-prefix "\t"))

(defun acme-indent-line ()
  "Auto-indent the current line."
  (interactive)
  (let* ((savep (point))
	 (indent (condition-case nil
		     (save-excursion
		       (forward-line 0)
		       (skip-chars-forward " \t")
		       (if (>= (point) savep) (setq savep nil))
		       (max (acme-calculate-indentation) 0))
		   (error 0))))
    (if savep
	(save-excursion (indent-line-to indent))
    (indent-line-to indent))))

(defun acme-indent-and-tab ()
  "Auto-indent and add tab"
  (interactive)
  (acme-indent-line)
  (insert "\t"))

(defun acme-anon-label (ch)
  "special handling for plus/minus at the beginning of line"
  (interactive)
  
  (let* ((line-string (buffer-substring-no-properties
                       (line-beginning-position) (line-end-position))))
    (if (string-match (concat "^[ \t]*\\(\\" ch "*\\)[ \t]*$") line-string)
        (progn (delete-region (line-beginning-position) (line-end-position))
               (beginning-of-line)
               (insert (match-string 1 line-string))
               (insert ch)
               (acme-indent-line)
               (insert "\t"))
      (call-interactively 'self-insert-command))))

(defun acme-calculate-indentation ()
  (or
   ;; Flush labels to the left margin.
   (and (looking-at "\\(\\sw\\|\\s_\\)+:") 0)
   ;; Flush labels to the left margin.
   (and (looking-at "[+-.@]") 0)
   ;; put !addr macro at column 0 to align with symbols
   (and (looking-at "!addr") 0)
   ;; Same thing for `;;;' comments.
   (and (looking-at "\\s<\\s<\\s<") 0)
   ;; Simple `;' comments go to the comment-column.
   (and (looking-at "\\s<\\(\\S<\\|\\'\\)") comment-column)
   ;; The rest goes at the first tab stop.
   (or (indent-next-tab-stop 0))))

(defun acme-colon ()
  "Insert a colon; if it follows a label, delete the label's indentation."
  (interactive)

  (if (not (nth 4 (syntax-ppss)))
      ;; check for long or short labels
      (let ((labelp nil) (longlabel nil))
        (if (>= (- (point) (line-beginning-position)) 8) (setq longlabel t))
        (save-excursion
          (skip-syntax-backward "w_.")
          (skip-syntax-backward " ")
          (if (setq labelp (bolp)) (delete-horizontal-space)))
        (call-interactively 'self-insert-command)
        (if longlabel
            (progn (delete-horizontal-space)
                   (insert "\n")
                   (tab-to-tab-stop))
          (when labelp
            (delete-horizontal-space)
            (tab-to-tab-stop))))

    ;; within comment just insert colon
    (call-interactively 'self-insert-command)))

;; Obsolete since Emacs-22.1.
(defalias 'acme-newline 'newline-and-indent)

(defun acme-comment ()
  "Convert an empty comment to a `larger' kind, or start a new one.
These are the known comment classes:

   1 -- comment to the right of the code (at the comment-column)
   2 -- comment on its own line, indented like code
   3 -- comment on its own line, beginning at the left-most column.

Suggested usage:  while writing your code, trigger acme-comment
repeatedly until you are satisfied with the kind of comment."
  (interactive)
  (comment-normalize-vars)
  (let (comempty comment)
    (save-excursion
      (beginning-of-line)
      (with-no-warnings
	(setq comment (comment-search-forward (line-end-position) t)))
      (setq comempty (looking-at "[ \t]*$")))

  (cond

   ;; just insert when within a string
   ;; ((nth 3 (syntax-ppss))
   ;;  (call-interactively 'self-insert-command))

   ;; Blank line?  Then start comment at code indent level.
   ;; Just like `comment-dwim'.  -stef
   ((save-excursion (beginning-of-line) (looking-at "^[ \t]*$"))
    (indent-according-to-mode)
    (insert acme-comment-char acme-comment-char ?\ ))

   ;; line as a comment but we are outside and before any non-whitespace char => comment line
   ((and comment (null (nth 4 (syntax-ppss))) (save-excursion (skip-chars-backward " \t") (bolp)))
    (comment-line 1))

   ;; Nonblank line w/o comment => start a comment at comment-column.
   ;; Also: point before the comment => jump inside.
   ;; ((and (null comment) (< (point) (line-end-position)))
   ;;  (comment-line 1))

   ;; before any non-whitespace char => comment line
   ((and (null comment) (< (point) (line-end-position)))
    (if (save-excursion (skip-chars-backward " \t") (bolp))
        (comment-line 1)
      (call-interactively 'self-insert-command)))
   ((and (null comment) (= (point) (line-end-position)))
    (indent-for-comment))

   ;; alternative to both ands above: always comment at the end of line
   ;; ((or (null comment) (< (point) comment))
   ;;  (indent-for-comment))

   ;; Flush-left or non-empty comment present => just insert character.
   ((or (not comempty) (save-excursion (goto-char comment) (bolp)))
    (insert acme-comment-char))

   ;; Empty code-level comment => upgrade to next comment level.
   ((save-excursion (goto-char comment) (skip-chars-backward " \t") (bolp))
    (goto-char comment)
    (insert acme-comment-char)
    (indent-for-comment))

   ;; Empty comment ends non-empty code line => new comment above.
   (t
    (goto-char comment)
    (skip-chars-backward " \t")
    (delete-region (point) (line-end-position))
    (beginning-of-line) (insert "\n") (backward-char)
    (acme-comment)))))

(provide 'acme-mode)

;;; acme-mode.el ends here
